/*RANDOM METHOD
MOPSO mobjective test functions
ZDT1 
PhD thesis of Sanaz Mostaghim

Last changes 30.10.2003

*/

#include<fstream>
#include<iostream>
#include<stdio.h>
#include<stdlib.h>
#include<string>
#include<vector>
#include<math.h>

using namespace std;


#include"mopso.h"

#define FALSE  0
#define TRUE   1

#define MAX_REP  80000 // maximum number of points in REPOSITORY (archive)

#define ASSERT(x)  if (!(x)) { cerr << "run time error: line " << __LINE__ <<\
                   " in file " << __FILE__ << "\n"; exit(1); }
#define ERROR(x)   cerr << "\nError: " << (x) << "\n"

#define OUTFILE     "mopso.out"      // file which records the run of the
                                     // selected algorithm


int noParameters = 30;  // dimension of x
int archivesize = 10000;
int IN = 0;
double W = 0.4;
double *low_range; // This defined to store the ranges of the parameters in the Parameter space
double *high_range;


/*---------------------------------------------------------------------------*/
void testfunction (double *x_ind, double *objectives)
{ 

// ZDT1 test function
	double  g = 0.0;
	for (int  i = 1; i < noParameters; i++)
	  g += x_ind[i];
	g = (g / (double(noParameters - 1) * 1.0)) * 9.0 + 1.0;
	double  h = 1.0 - sqrt(x_ind[0] / g);
	objectives[0] = x_ind[0];
	objectives[1] = g * h; 
 
  
} /* testfunction */  


/*---------------------------------------------------------------------------*/
// random number generation 
/*---------------------------------------------------------------------------*/
int  RandomInt(int  min, int  max)
  // generates a random integer 
{
  return (int) (min + (rand() % (max - min + 1)));
} /* RandomInt */

/*---------------------------------------------------------------------------*/
double  RandomDouble(double  min, double  max)
  // generates a random floating point number 
{
  return (min + ((double(rand()) / double(RAND_MAX)) * (max - min)));
} /* RandomDouble */

/*---------------------------------------------------------------------------*/
// reading Parameters
/*---------------------------------------------------------------------------*/
void  ReadParameters(int  argc, char*  argv[], globals&  params)
  // set parameters according to command line specification
{
  if (argc < 2) {
    ERROR("Missing Arguments");
    cout << "\nUsage: 1) Guide Algorithm (for this version just:  random )" <<
      "\n       2) Maximum Number of Cycles (150)" <<
      "\n       3) Population Size (100)" <<
      "\n       4) Archive size (50)" <<
      "\n       5) Number of Parameters (30)" <<
      "\n       6) Number of Objectives (2)" <<
      "\n       7) Turbulence factor (0.1)" <<
      "\n 	8) W (Inertia weight (0.4))" <<
      "\n 	9) IN (0 : no input archive, 1: read from input archive)" <<
      "\nAt least the first argument has to be specified.\n\n";
    exit(1);
  }
  if (strcmp("random", argv[1]) == 0)       params.findguide = RANDOM;
  else                                   ERROR("Unknown algorithm");  
  if (argc > 2)  sscanf(argv[2], "%d", &(params.maxCycles));
  else           params.maxCycles = 150;
  if (argc > 3)  sscanf(argv[3], "%d", &(params.popSize));
  else           params.popSize = 100;
  if (argc > 4)  sscanf(argv[4], "%d", &(params.REPSize));
  else           params.REPSize = 50;
  if (argc > 5)  sscanf(argv[5], "%d", &(noParameters));
  else           noParameters = 30;
  if (argc > 6)  sscanf(argv[6], "%d", &(params.noObjectives));
  else           params.noObjectives = 2;
  if (argc > 7)  sscanf(argv[7], "%lf", &(params.mut));
  else           params.mut = 0.1;
  if (argc > 8)  sscanf(argv[8], "%lf", &(W));
  else           W = 0.4;
  if (argc > 9)  sscanf(argv[9], "%d", &(IN));
  else           IN = 0;
 
  params.popSeed = 1523;
  params.dataSeed = 1825;

  low_range = new double[noParameters];
  high_range = new double[noParameters];
  for (int i = 0; i < noParameters; i++){
  	low_range[i] = 0;
	high_range[i] = 1;
  }
} /*ReadParameters*/

/*---------------------------------------------------------------------------*/
int ReadREP(globals&  params, population*  REP){
  int size;
  int i,j; 
  double x;
  ifstream  inFiletarget("archive.in");
  if (!inFiletarget) {
    ERROR("Cannot open input file: archive.in");
    exit(1);
  }
  
  inFiletarget >> size;
  REP->size = size;
  for (int j = 0; j< REP->size; j++){
	for (int i = 0; i <noParameters; i++){
		inFiletarget >> x;
		REP->indArray[j].SetPosition(x,i);
		REP->indArray[j].SetVelocity(0,i);
	}
  }
  inFiletarget.close();
  
  return size;
} /*ReadREP*/

/*----- test ------------------------------------------------------------*/
test::test(int noobjective, int noparameters)
  // the data set is created here
{ 

    noParameters = noparameters; 
    noObjectives = noobjective;  
    cout << "noParameters: " <<noParameters << "   noObjectives :" <<noObjectives<<endl;
  
} /* test::test */

/*----- individual ----------------------------------------------------------*/
test*          individual::testPtr;
/*---------------------------------------------------------------------------*/
individual::individual(void)
  // creates a new individual (default constructor)
{
  ASSERT (testPtr != NULL);

  objectives = new double[testPtr->noObjectives];
  x_ind = new double[testPtr->noParameters]; // position x
  v_ind = new double[testPtr->noParameters]; // velocity
  
 for (int i= 0; i<noParameters; i++){ // this loop initializes the x and v
  	x_ind[i] = RandomDouble(low_range[i],high_range[i]); 
  	v_ind[i] = RandomDouble(low_range[i],high_range[i]);
 }

  objValid = FALSE;
  fitValid = FALSE;
} /* individual::individual */
/*---------------------------------------------------------------------------*/
individual::~individual(void)
{
  if (objectives != NULL) delete [] objectives;
  if (x_ind != NULL) delete []  x_ind;
  if (v_ind != NULL) delete []  v_ind;
  
} /* individual::~individual */
/*---------------------------------------------------------------------------*/
void  individual::Reset(test*  testData)
  // static member function to set initial parameters
{
  ASSERT(testData != NULL);

  testPtr = testData;

} /* individual::Reset */
/*---------------------------------------------------------------------------*/
inline double  individual::GetObjective(int  no)
{

  ASSERT(no >= 0 && no < testPtr->noObjectives);
  if (!objValid) {
 	
	testfunction(x_ind, objectives);
	
	objValid = TRUE;
    }
 
  return objectives[no];
  
} /* individual::GetObjective */
/*---------------------------------------------------------------------------*/
void individual::SetVelocity(double v, int j){
	v_ind[j]= v;
	objValid = FALSE;
} /* individual::SetVelocity */
/*---------------------------------------------------------------------------*/
void individual::SetPosition(double x, int j){
	x_ind[j]= x;

	if (x < low_range[j])
		x_ind[j]= low_range[j]; 
	else if (x > high_range[j])
		x_ind[j]= high_range[j]; 
		
	objValid = FALSE;
} /* individual::SetPosition */
/*---------------------------------------------------------------------------*/
double  individual::GetVelocity(int j){
	return v_ind[j];
	
} /* individual::GetVelocity */
/*---------------------------------------------------------------------------*/
double  individual::GetPosition(int j){
	return x_ind[j];
} /* individual::GetPosition */
/*---------------------------------------------------------------------------*/
int  individual::Dominates(individual&  ind, int  noConsideredObjectives)
{
    short  greaterInOneObjective = FALSE;
    int k;
    for ( k = 0; k < noConsideredObjectives && GetObjective(k) <= ind.GetObjective(k); k++)
      if (GetObjective(k) < ind.GetObjective(k))  greaterInOneObjective = TRUE;
    if (k >= noConsideredObjectives && greaterInOneObjective) 
      return TRUE;
    
    return FALSE;
  
} /* individual::Dominates */
/*---------------------------------------------------------------------------*/

int  individual::Dominates(individual&  ind)
  // returns TRUE if the individual dominates the specified one 'ind'
{
  return Dominates(ind, testPtr->noObjectives);
} /* individual::Dominates */
/*---------------------------------------------------------------------------*/
void  CopyIndividual(individual&  indSrc, individual&  indDest)
  // the contents of individual 'indSrc' are copied to individual 'indDest'
{
  int i;
  indDest.objValid = indSrc.objValid;
  indDest.fitness = indSrc.fitness;
  indDest.fitValid = indSrc.fitValid;

  for ( i = 0; i < indSrc.testPtr->noParameters; i++){
    indDest.x_ind[i] = indSrc.x_ind[i];
    indDest.v_ind[i] = indSrc.v_ind[i];
  }
  ASSERT(indSrc.objectives != NULL && indDest.objectives != NULL);
  for (i = 0; i < indSrc.testPtr->noObjectives; i++)
    indDest.objectives[i] = indSrc.objectives[i];
} /* CopyIndividual */

/*----- population ----------------------------------------------------------*/
test*  population::testPtr;

population::population(int  popSize, test*  testData)
{

  ASSERT(popSize > 0);
  size = popSize;
  maxSize = popSize;
  indArray = new individual[popSize];
  ASSERT(indArray != NULL);
  
  TempParetoSet = new individual[popSize];
  ASSERT(TempParetoSet != NULL);
  

  tempArray = new individualPtr[popSize];
  ASSERT(tempArray != NULL);

  pbest = new individual[popSize]; // pbest
  for (int i =0; i<size; i++){ //pbest init
  	for (int j = 0; j<noParameters; j++){
	  	pbest[i].SetPosition(indArray[i].GetPosition(j),j);
		pbest[i].SetVelocity(indArray[i].GetVelocity(j),j);
	}
  }
  ASSERT(testData != NULL);
  testPtr = testData;
  
  
  clusters = new int[maxSize];
  clusterList = new int[maxSize];
  selectedSolutions = new int[maxSize];
  ASSERT(clusters != NULL && clusterList != NULL && selectedSolutions != NULL);
  
} /* population::population */
/*---------------------------------------------------------------------------*/
population::~population(void)
{
  if (indArray != NULL)  delete [] indArray;
  if (tempArray != NULL)  delete [] tempArray;
  if (clusters != NULL)  delete [] clusters;
  if (clusterList != NULL)  delete [] clusterList;
  if (selectedSolutions != NULL)  delete [] selectedSolutions;  
  
} /* population::~population */
/*---------------------------------------------------------------------------*/
void  population::Print(ostream&  out)
{
   
  int i,j;
//  out << "Objective Values :\n";
  for (i = 0; i < size; i++) {
      for (j = 0; j < testPtr->noObjectives; j++) {
	out << indArray[i].GetObjective(j);
	out << "\t ";
      }
      out << "\n ";
  }
  out << "\n\n";
/*  out << "Parameter Values :\n";
  for (i = 0; i < size; i++) {
        for (j = 0; j < noParameters; j++) {
		out <<indArray[i].GetPosition(j);     
                out << "\t ";
      }
      out << "\n ";
  }
  out << "\n ";
 */ 
  out.flush();
} /* population::Print */

/*---------------------------------------------------------------------------*/
void  population::GetREP(population*  paretoSet)
{
  // determine Pareto points within population; 'tempArray[i] == NULL' means
  // point i is not dominated
  ASSERT(size <= maxSize);
  int i,j;
  for (i = 0; i < size; i++)  tempArray[i] = NULL;
  for (i = 0; i < size; i++)
    for (int  j = i + 1; j < size; j++)
      if (indArray[i].Dominates(indArray[j]))  tempArray[j] = indArray;
      else if (indArray[j].Dominates(indArray[i]))  tempArray[i] = indArray;
  // combine pareto points found with points in 'paretoSet'
  ASSERT(paretoSet->size <= paretoSet->maxSize && paretoSet->size >= 0);
  i = 0;
  while (i < size) {
    if (tempArray[i] == NULL) {
      // check whether point is already in pareto set
      short  iIsDominated = FALSE;
      for (j = 0; j < paretoSet->size; j++) {
	if (indArray[i].Dominates(paretoSet->indArray[j]))  break;
	else if (paretoSet->indArray[j].Dominates(indArray[i])) {
	  iIsDominated = TRUE;
	  break;
	}
      }
      if (!iIsDominated) {
	// check whether other points in pareto set are dominated by the
	// point actually considered
	if (j == paretoSet->maxSize) // 
		j--;
	ASSERT(j < paretoSet->maxSize);
	
	CopyIndividual(indArray[i], paretoSet->indArray[j]);
	if (j < paretoSet->size) {
	  j++;
	  while (j < paretoSet->size) 
	    if (indArray[i].Dominates(paretoSet->indArray[j])) {
	      paretoSet->size--;
	      CopyIndividual(paretoSet->indArray[paretoSet->size],
			     paretoSet->indArray[j]);
	    }
	    else
	      j++;
	}
	else  paretoSet->size++;
      }
    }
    i++;
  }

} /* population::GetREP */
/*---------------------------------------------------------------------------*/
// Clustering
/*---------------------------------------------------------------------------*/
double  IndDistance(individual&  ind1, individual&  ind2)
  // phenotypic distance regarding the objective values
{
  double phen_dist = 0;
  for (int i = 0; i< ind1.testPtr->noObjectives; i++)
  	phen_dist = (ind1.GetObjective(i)-ind2.GetObjective(i))*(ind1.GetObjective(i)-ind2.GetObjective(i)) + phen_dist;
	
  phen_dist = sqrt(phen_dist);	
  return (phen_dist);

} /* PhenDistObjectives */

/*---------------------------------------------------------------------------*/

double  population::ClusterDistance(int  cluster1, int  cluster2)
{
  double  sum = 0;
  int  numOfPairs = 0;
  int  c1 = clusters[cluster1];
  while (c1 >= 0) {
    int  c2 = clusters[cluster2];
    while (c2 >= 0) {
      sum += IndDistance(indArray[c1], indArray[c2]);
      numOfPairs++;
      c2 = clusterList[c2];
    }
    c1 = clusterList[c1];
  }
  return (sum / double(numOfPairs));
  
} /* population::ClusterDistance */
/*---------------------------------------------------------------------------*/

void  population::JoinClusters(int  cluster1, int  cluster2, int&  numOfClusters)
{
  ASSERT(clusters[cluster1] >= 0);
  int  c1 = clusters[cluster1];
  while (clusterList[c1] >= 0)
    c1 = clusterList[c1];
  clusterList[c1] = clusters[cluster2];
  numOfClusters--;
  clusters[cluster2] = clusters[numOfClusters];
} /* population::JoinClusters */
/*---------------------------------------------------------------------------*/
 
int  population::ClusterCentroid(int  cluster)
{
  double  minSum = -1;
  int  minIndex;
  int  c1 = clusters[cluster];
  while (c1 >= 0) {
    int  c2 = clusters[cluster];
    double  sum = 0;
    while (c2 >= 0) {
      sum += IndDistance(indArray[c1], indArray[c2]);
      c2 = clusterList[c2];
    }
    if (sum < minSum || minSum < 0) {
      minSum = sum;
      minIndex = c1;
    }
    c1 = clusterList[c1];
  }
  return minIndex;
  
} /* population::ClusterCentroid */
/*---------------------------------------------------------------------------*/

void  population::ReduceByClustering(int  reducedSize)
  // reduce set of individuals by clustering using specified distance metric
{
  
  if (size <= reducedSize || reducedSize < 1)  return;
  // firstly, define one-element-clusters
  int  numOfClusters = size;
  int i;
  for ( i = 0 ; i < size; i++) {
    clusters[i] = i;
    clusterList[i] = -1;
  }
  // join clusters using average linkage method
  while (numOfClusters > reducedSize) {
    // find cluster pair with minimal distance
    double  minDist = -1;
    int     join1, join2;
    int c,d;
    for (  c = 0; c < numOfClusters; c++) {
      for (  d = c + 1; d < numOfClusters; d++) {
	double  dist = ClusterDistance(c, d);
	if (dist < minDist || minDist < 0) {
	  minDist = dist;
	  join1 = c;
	  join2 = d;
	}
      }
    }
    // join cluster pair
    JoinClusters(join1, join2, numOfClusters);
  }
  // chose for each cluster representative solution (centroid)
  for (i = 0; i < size; i++)
    selectedSolutions[i] = FALSE;
  for (i = 0; i < numOfClusters; i++)
    selectedSolutions[ClusterCentroid(i)] = TRUE;
  int  counter = 0;
  for (i = 0; i < size; i++) {
    if (selectedSolutions[i]) {
      individual  temp;
      CopyIndividual(indArray[i], temp);
      CopyIndividual(indArray[counter], indArray[i]);
      CopyIndividual(temp, indArray[counter]);
      counter++;
    }
  }
  size = counter;
} /* population::ReduceByClustering */

/*---------------------------------------------------------------------------*/
double findMax(individual *  Array, int size, int noObjective){
	double temp = Array[0].GetObjective(noObjective);
	for (int i = 1; i < size; i++){
		if (Array[i].GetObjective(noObjective)> temp){
			temp = Array[i].GetObjective(noObjective);
		}
	}
	return (temp);
} /* findMax */
/*---------------------------------------------------------------------------*/
void  population::Update(population*  rep, globals&  params)
{
  double v,x;
  int i,j,k,h;
  
  
  /*--------------------------------------------------------------------------*/
  // main loop on the population
  for ( i= 0;i < size; i++){
	if (indArray[i].Dominates(pbest[i])){ // updating pbest	
		for ( k = 0; k<noParameters; k++)
			pbest[i].SetPosition(indArray[i].GetPosition(k), k);
	}
	
	// now to find the global guide reph (Here I do the random method)
        /*--------------------------------------------------------------------------*/
	individual reph;
	int selectfromRep = RandomInt(0, rep->size-1);
	
	CopyIndividual(rep->indArray[selectfromRep], reph);
	/*----------------------------------------------------------------------------*/		
	// now we have reph (the global guide)
	for ( j = 0; j< noParameters ; j++){ 
		v = indArray[i].GetVelocity(j) * W + 
		    RandomDouble(0,1) * (pbest[i].GetPosition(j) - indArray[i].GetPosition(j)) +
		    RandomDouble(0,1) * (reph.GetPosition(j) - indArray[i].GetPosition(j));
		

		indArray[i].SetVelocity(v,j);
		
		x = indArray[i].GetVelocity(j)+ indArray[i].GetPosition(j);

		indArray[i].SetPosition(x,j);
	}
	
	// mutation
	double myprobability = RandomDouble(0, 1);
	if ((myprobability < params.mut) && (params.mut != 0)){
		for ( j = 0; j < 9 ; j++){
			x = indArray[i].GetPosition(j) + RandomDouble(-1,1) * indArray[i].GetPosition(j); 
			indArray[i].SetPosition(x,j);
		}
		
	}
		
	
	} // end of main loop

} /* population::Update */
/*---------------------------------------------------------------------------*/
void  population::Evaluate(void)
{
	for ( int i= 0;i < size; i++){
		indArray[i].GetObjective(1);
	}
} /* population::Evaluate */
/*----- main ----------------------------------------------------------------*/

int  main(int  argc, char*  argv[])
{

  globals  params;
  ReadParameters(argc, argv, params);
  srand(params.dataSeed);
  test*  testPtr = new test(params.noObjectives, noParameters);
  individual::Reset(testPtr);
  
  cout << "params.maxCycles: "<<params.maxCycles<<endl;
  cout << "params.pop: "<<params.popSize<<endl;
  cout << "params.pareto: "<<params.REPSize<<endl;
  cout << "params.param: "<<noParameters<<endl;
  cout << "Turbulence factor: "<<params.mut<<endl;
  cout << "params.popseed: "<<params.popSeed<<endl;
  cout << "params.dataseed: "<<params.dataSeed<<endl;
  cout << "params.obj: "<<params.noObjectives<<endl;
  
  cout << endl;
  cout << endl;

  
  // prepare output streams
  ofstream  outFile(OUTFILE);
  if (!outFile) {
    ERROR("Cannot open output file");
    exit(1);
  }
  outFile.setf(ios::fixed, ios::floatfield);
  outFile.precision(10);

  
  // init populations
  srand(params.popSeed);
  population* pop = new population(params.popSize, testPtr); // new population 
  population*  REP = new population(MAX_REP, testPtr); // new REPOSITORY (archive)
  REP->size = 0;
  
  if (IN == 1){
	  cout <<" YOU HAVE ACTIVATED THE INPUT OPTION...  THE INPUT ARCHIVE MUST BE IN archive.in FILE" << endl;
	  // read the input rep from the file archive.in
	  REP->size = ReadREP(params, REP);
	  cout << "Input rep size: "<<REP->size << endl;
   }
 
  pop->Evaluate();
  // first we have to Evaluate the population and make the REP
  pop->GetREP(REP); // it must be here; don't delete this line
  if (REP->size > params.REPSize)
  	REP->ReduceByClustering(params.REPSize); 

  // main loop
  int  cycles = 0;
  do {
    cout <<cycles <<" -"<< "\t";
    pop->GetREP(REP);
    
    if (REP->size >= params.REPSize)
 	REP->ReduceByClustering(params.REPSize); 
     
    if (cycles == params.maxCycles)	REP->Print(outFile);
    
    pop->Update(REP, params);
    
    cycles++;
    cout <<"rep size: "<<REP->size << " \n ";  
  } while (cycles <= params.maxCycles);
 
 
  delete [] low_range;
  delete [] high_range;
  outFile.close();
} /* main */






